#!/usr/bin/env python
#
# AutoReply2 automatic mail responder, v2.0b
# (c) 2005, T. Morstein <tmr@fooboo.org>
#
# TODO:
#  * rulesets implementation
#  * implementation of more than Maildir
#  * remote mailbox access through POP3
#############################################################################

import os
import sys
import sha
import socket
import string
import logging
import mailbox
import smtplib
import email.MIMEText
import ConfigParser


#############################################################################
# 1. Functions
# ------------
# {{{

#############################################################################
# 1.1. Config: loadConfig(), checkConfig(), isString(), isLevel(),
#              isNumber(), isStatus(), isMbxType(), isInternalFile(),
#              isLogFmt(), isProfile()
# -------------------------------------------------------------------
# TODO: isString(), isNumber()
# {{{

#############################################################################
# 1.1.1. loadConfig( file )
# -------------------------
# Load configuration from file;
# return False on problems, else return configuration structure
# {{{
def loadConfig( file ):

  # Is file variable really regural config file ?
  if not os.path.isfile( file ):
    logger.warn( 'file cannot be found: \'%s\'' % file )
    return False

  cobj = ConfigParser.ConfigParser()
  cobj.read( file )

  cfg = {}
  for section in cobj.sections():
    cfg[ section ] = {}
    for option in cobj.options( section ):
      cfg[ section ][ option ] = cobj.get( section, option )

  # Return configuration in form 'dict-in-dict'
  return cfg

# }}}

#############################################################################
# 1.1.2. isString( value )
# ------------------------
# Return True if specified value is string
# {{{
def isString( value ):
  return True

# }}}

#############################################################################
# 1.1.3. isLevel( value )
# -----------------------
# Return True if specified value is one of loglevels
# {{{
def isLevel( value ):

  levels   = [ 'debug', 'info',  'warn', 'warning',
               'error', 'fatal', 'critical' ]

  if value in levels:
    return True
  else:
    return False

# }}}

#############################################################################
# 1.1.4. isNumber( value )
# ------------------------
# Return True if specified value is number
# {{{
def isNumber( value ):
  return True

# }}}

#############################################################################
# 1.1.5. isStatus( value )
# ------------------------
# Return True if value is status - enabled or disabled
# {{{
def isStatus( value ):

  if value in [ 'enabled', 'disabled' ]:
    return True
  else:
    return False

# }}}

#############################################################################
# 1.1.6. isMbxType( value )
# -------------------------
# Return True if value si supported mailbox type
# {{{
def isMbxType( value ):

  mbxTypes = [ 'pop', 'maildir' ]

  if string.lower( value ) in mbxTypes:
    return True
  else:
    return False

# }}}

#############################################################################
# 1.1.7. isInternalFile( value )
# ------------------------------
# Check if value is existing file or if we have chance to create this file
# {{{
def isInternalFile( value ):

  if os.path.isfile( value ):
    if not os.access( value, os.W_OK ):
      logger.warning( 'bad permissions on specified file' )
      return False
  else:
    dst = os.path.dirname( value )
    if not os.path.isdir( dst ):
      logger.warning( 'file\'s directory cannot be found' )
      return False
    elif not os.access( dst, os.W_OK ):
      logger.warning( 'bad permissions in file\'s directory' )

  return True

# }}}

#############################################################################
# 1.1.8. isLogFmt( value )
# ------------------------
# Check if value is a valid log format: 'vms' or 'unix' style
# {{{
def isLogFmt( value ):

  if string.lower( value ) not in [ 'vms', 'unix' ]:
    return False

  return True

# }}}

#############################################################################
# 1.1.9. isProfile( value )
# -------------------------
# Return True if value is a valid user profile source
# {{{
def isProfile( value ):

  if value not in [ 'system', 'defined' ]:
    return False

  return True

# }}}

#############################################################################
# 1.1.10. checkConfig( cfsct )
# ---------------------------
# Check configuration values and dependencies;
# if there are mistakes in configuration return False, else return True
# XXX: required values and dependencies checking
# {{{
def checkConfig( cfsct ):

  # if logfile doesn't exist, logger create it
  logging  = { 'logfile':    isInternalFile,        \
               'loglevel':   isLevel,               \
               'logformat':  isLogFmt               \
             }
  network  = { 'smtpserver': isString,              \
               'smtpauth':   isStatus,              \
               'smtptls':    isStatus,              \
               'smtpport':   isNumber,              \
               'smtplogin':  isString,              \
               'smtppasswd': isString,              \
               'popserver':  isString,              \
               'popssl':     isStatus,              \
               'poplogin':   isString,              \
               'poppasswd':  isString               \
             }
  mailbox  = { 'mboxtype':   isMbxType,             \
               # XXX: until we are supporting only Maildir:
               'mboxpath':   os.path.isdir,         \
               # This will be AR2's home directory
               # for temporary messages and this database
               'dbpath':     isInternalFile         \
             }
  personal = { 'profile':    isProfile,             \
               'fullname':   isString,              \
               'mailaddr':   isString,              \
               'announce':   os.path.isfile         \
             }
  # XXX: RULES
  # rules  = { personal[ 'mailaddr' ]: 'deny'       \
  #          }
  all      = { 'logging':  logging,     \
               'network':  network,     \
               'mailbox':  mailbox,     \
               'personal': personal     \
  #          , 'rules':    rules        \
             }
#  basReq   = { 'logfile':     {},                     \
#               'loglevel':    {},                     \
#               'logformat':   {},                     \
#               'smtpserver':  {},                     \
#               'announce':    {},                     \
#               'mboxtype':    {                       \
#                        'maildir':  [ 'mboxpath' ],   \
#                        'pop':      [ 'popserver',    \
#                                      'poplogin',     \
#                                      'poppasswd' ]   \
#               },                                     \
#               'profile':     {                       \
#                        'defined':  [ 'fullename',    \
#                                      'mailaddr' ],   \
#                        'system':   []                \
#               }                                      \
#             }
#  optReq   = { 'smtpauth':    {                       \
#                        'enabled':  [ 'smtplogin',    \
#                                      'smtppasswd' ], \
#                        'disabled': []                \
#               },                                     \
#               'smtptls':     {},                     \
#               'smtpport':    {}                      \
#             }

  # Sections, keys and values checks
  for curs in cfsct.keys():
    if curs not in all.keys():
      logger.warn( 'unknown section: \'%s\'' % curs )
    else:
      for curk in cfsct[ curs ].keys():
        if curk not in all[ curs ].keys():
          logger.warn( 'unknown key \'%s\' in section \'%s\'' % \
              ( curk, curs ) )
        else:
          if not ( all[ curs ][ curk ] )( cfsct[ curs ][ curk ] ):
            logger.warn( 'invalid value for key \'%s\'' \
                ' in section \'%s\'' % ( curk, curs ) )
            return False
          else:
            logger.debug( 'section \'%s\', key \'%s\'' \
                ' everything is ok' % ( curs, curk ) )

  # Checks for required values and dependencies
  # XXX: MISSING !!
  # Try to combine this with previous block

  return True

# }}}

# }}}

#############################################################################
# 1.2. Logging: initLogger()
# --------------------------
# {{{

#############################################################################
# 1.2.1. initLogger( dst, type, level, fmt )
# ------------------------------------------
# Initialize logger for specified dst (stream or file) and loglevel
# Return set of two or three values; if the first is True, second is logger;
# when the first is False, second is error message and third is handler
# {{{
def initLogger( dst, type, level, fmt ):

  handlers  = { 'stream':   logging.StreamHandler, \
                'file':     logging.FileHandler    \
              }
  type      = string.lower( type )
  loglevels = { 'debug':    logging.DEBUG,         \
                'info':     logging.INFO,          \
                'warn':     logging.WARNING,       \
                'warning':  logging.WARNING,       \
                'error':    logging.ERROR,         \
                'fatal':    logging.FATAL,         \
                'critical': logging.CRITICAL       \
              }
  if type not in handlers.keys():
    return [ False, 'unsupported handler type !' ]
  if level not in loglevels.keys():
    return [ False, 'unsupported logging level !' ]

  try:
    lgr       = logging.getLogger( 'AutoReply2' )
    hdlr      = ( handlers[ type ] )( dst )
    fmt = string.lower( fmt )
    if fmt == 'unix':
      formatter = logging.Formatter(                       \
          '%(asctime)s: %(filename)s: [%(levelname)s]: %(message)s' )
    elif fmt == 'vms':
      formatter = logging.Formatter( \
'%%%%%%%%%%%%%%%%%%%%%% %(asctime)s %%%%%%%%%%%%%%%%%%%%%%\n\
Module: %(module)s\n\
Severity: %(levelname)s\n\
Line: %(lineno)d\n\
Message: %(message)s\n\
' )
    else:
      return [ False, 'Unknown log format !' ]
    hdlr.setFormatter( formatter )
    lgr.addHandler( hdlr )
    lgr.setLevel( loglevels[ level ] )
  except IOError, errMsg:
    return [ False, errMsg ]

  return [ True, lgr, hdlr ]

# }}}

#############################################################################
# 1.2.2. reInitLogger( dst, type, level, fmt )
# --------------------------------------------
# Disable old logger and initialize new one
# {{{
def reInitLogger( dst, type, level, hldr, fmt ):

  # Stop old logger
  logger.removeHandler( hldr )

  # Init new logger
  return initLogger( dst, type, level, fmt )

# }}}

# }}}

#############################################################################
# 1.3. Core: mboxOpen(), loadDatabase(), refreshDatabase(),
#            msgHash(), buildQueue(), checkMsgRules(), applyPolicies()
# --------------------------------------------------------------------
# TODO: applyPolicies()
# {{{

#############################################################################
# 1.3.1. mboxOpen( mbxCfg )
# -------------------------
# Open mailbox specified in 'mailbox' section of configuration;
# return mailbox.<mboxType> instance
# {{{
def mboxOpen( mbxCfg ):

  types = { 'Maildir': mailbox.Maildir }

  if mbxCfg[ 'mboxtype' ] in types.keys():
    try:
      return ( types[ mbxCfg[ 'mboxtype' ] ] )( mbxCfg[ 'mboxpath' ] )
    except OSError, errMsg:
      logger.error( 'cannot open mailbox: %s' % errMsg )

  return False

# }}}

#############################################################################
# 1.3.2. loadDatabase( db )
# -------------------------
# Load database file with already processed messages and return list of them;
# on problems return False
# {{{
def loadDatabase( db ):

  res = []

  if not os.path.isfile( db ):
    logger.debug( 'database cannot found, creating new one' )
    f = open( db, 'w' )
    f.close()
    return res

  try:
    f = open( db, 'r' )
  except IOError, errMsg:
    logger.warning( 'cannot open (r) database: %s' % errMsg )
    return False

  while True:
    rec = f.readline()
    if not rec:
      break
    res.append( rec.strip() )

  f.close()

  return res

# }}}

#############################################################################
# 1.3.3. refreshDatabase( db, proc )
# ----------------------------------
# Refresh database file with already processed messages and
# return True if OK; on problems return False
# {{{
def refreshDatabase( db, proc ):

  dbNew = '%s.%s' % ( db, os.getpid() )
  dbOld = db + '.old'

  try:
    f = open( dbNew, 'w' )
  except IOError, errMsg:
    logger.warning( 'cannot open (w) database: %s' % errMsg )

  for cur in proc:
    f.write( '%s\n' % cur )
    f.flush()

  f.close()

  try:
    os.link( db, dbOld )
    os.unlink( db )
  except OSError, errMsg:
    logger.warn( 'cannot backup old database: %s' % errMsg )

  try:
    os.link( dbNew, db )
  except OSError, errMsg:
    logger.warn( 'cannot link new database: %s' % errMsg )
    logger.warn( 'trying to get back o;d database' )
    try:
      os.link( dbOld, db )
    except OSError, errMsg:
      logger.warn( 'cannot get back old database' )
    logger.warn( 'there were problems with updating database' )
    return False

  os.unlink( dbOld )
  os.unlink( dbNew )

  logger.info( 'database successfully updated' )

  return True

# }}}

#############################################################################
# 1.3.4. msgHash( msg )
# ---------------------
# Return SHA1 hash of rtc822 'msg' message
# {{{
def msgHash( msg ):

  return sha.new( getText( msg.fp.name ) ).hexdigest()

# }}}

#############################################################################
# 1.3.5. checkMsgRules( msg )
# ---------------------------
# Return True if message is allowed to process else return False
# {{{
def checkMsgRules( msg ):

  hdrsBad = [ 'list-id' ]

  for hdr in hdrsBad:
    if msg.has_key( hdr ):
      return False

  return True

# }}}

#############################################################################
# 1.3.6. applyPolicies( queue, rules )
# ------------------------------------
# Reduce specified queue on the basis of configured rules;
# return modification of input queue
# {{{
def applyPolicies( queue, rules ):

  # XXX: parsing rules, reducing queue

  return queue

# }}}

#############################################################################
# 1.3.7. buildQueue( mbx, prefix, db, rules )
# -------------------------------------------
# Build queue in format <subject>:<adress> and
# update database of processed messages;
# return list of adresses or in troubles return False
# {{{
def buildQueue( mbx, prefix, db, rules ):

  # Headers containing adress of recipient
  rcptHdrs  = [ 'reply-to', 'return-path', 'from', 'sender' ]

  # Queue containing jobs to process in format <subject>:<address>
  queue     = []

  # Hashes of already processed messages
  processedMsgs = loadDatabase( db )
  if processedMsgs == False:
    logger.warn( 'cannot load database of processed messages' )
    return False

  prefLen = len( prefix )

  while True:
    curMsg = mbx.next()
    if not curMsg:
      break
    curHash = msgHash( curMsg )
    if curHash in processedMsgs:
      logger.debug( 'message (%s) already processed, skipping' % curHash )
    else:
      # Select only new mails
      if curMsg.fp.name[ :prefLen ] == prefix:
        if checkMsgRules( curMsg ):
          for hdr in rcptHdrs:
            if curMsg.has_key( hdr ):
              contact = curMsg.getaddr( hdr )[ 1 ]
              if curMsg.has_key( 'subject' ):
                subject = curMsg.getheader( 'subject' )
              else:
                subject = '<no subject>'
              queue.append( [ subject, contact ] )
              processedMsgs.append( msgHash( curMsg ) )
              logger.info( 'collected new message (%s) from \'%s\'' % \
                  ( curHash, contact ) )
              break
        else:
          logger.debug( 'message (%s) disabled by ruleset' % curHash )

  if not refreshDatabase( db, processedMsgs ):
    logger.warning( 'problem with refreshing database' )
    return False

  return applyPolicies( dict( queue ), rules )

# }}}

#############################################################################
# 1.3.8. sendMsg( rcpt, subj, body, netcf, perscf )
# -------------------------------------------------
# Send mail message
# {{{
def sendMsg( rcpt, subj, body, netcf, perscf ):

  try:
    server = smtplib.SMTP( netcf[ 'smtpserver' ] )
  except (socket.error, socket.gaierror), errMsg:
    raise smtplib.SMTPException, errMsg[ 1 ]

  if netcf[ 'smtpauth' ] == 'enabled':
    try:
      server.login( netcf[ 'smtplogin' ], netcf[ 'smtppasswd' ] )
    except (smtplib.SMTPError, smtplib.SMTPAuthenticationError ), errMsg:
      raise smtplib.SMTPException, errMsg
  if netcf[ 'smtptls' ] == 'enabled':
    server.starttls()

  msg = email.MIMEText.MIMEText( body )
  msg[ 'To' ] = rcpt
  msg[ 'Subject' ] = 'Re: %s' % subj
  msg[ 'From' ] = '"%s" <%s>' % ( perscf[ 'fullname' ], perscf[ 'mailaddr' ] )
  msg[ 'X-AutoReply2-Version' ]  = __version__
  msg[ 'X-AutoReply2-Homepage' ] = __homepage__

  smtpresult = server.sendmail( 'AutoReply2', \
                                rcpt, msg.as_string() )
  if smtpresult:
    errMsg = ''
    for cur in smtpresult.keys():
      errMsg = 'Could not delivery mail to: %s; server said: %s %s %s' % \
          ( cur, smtpresult[ cur ][ 0 ], smtpresult[ cur ][ 1 ], errMsg )
    raise smtplib.SMTPException, errMsg

  server.quit()

  return True

# }}}

#############################################################################
# 1.3.9. getText( file )
# ----------------------
# Return text from file
# {{{
def getText( file ):

  message = ''

  f = open( file, 'r' )

  while True:
    ln = f.readline()
    if not ln:
      break
    message = message + ln

  return message

# }}}

# }}}

# }}}

#############################################################################
# 2. Main code
# ------------
# {{{

# Some informations
__version__ = '2.0b-cvs, $Id: autoreply2.py,v 1.10 2005/12/25 19:19:06 tmr Exp $'
__homepage__ = 'http://www.fooboo.org/autoreply2/'

# Initialize stdout logger
# Choose 'debug' to enable higher level of verbosity
# while reading configuration
# Change 'unix' to 'vms' if you want to init
# this logger with vms logging style
res = initLogger( sys.stdout, 'stream', 'info', 'unix' )
if res[ 0 ]:
  logger = res[ 1 ]
else:
  print '%%AUTOREPLY2-W-LOGGER, cannot init stdout logger\n \\%s\\' % res[ 1 ]
  sys.exit( 1 )

# Load configuration
config = loadConfig( 'autoreply.cfg' )
if not config:
  logger.error( 'cannot load configuration' )
  sys.exit( 1 )

# Validate configuration
if not checkConfig( config ):
  logger.error( 'configuration check failed' )
  sys.exit( 1 )

# Initialize logging specified by configuration
res = reInitLogger( config[ 'logging' ][ 'logfile' ], 'file',     \
                    config[ 'logging' ][ 'loglevel' ], res[ 2 ],  \
                    config[ 'logging' ][ 'logformat' ] )
if res[ 0 ]:
  logger = res[ 1 ]
else:
  print '%%AUTOREPLY2-W-LOGGER, cannot init file logger\n \\%s\\' % res[ 1 ]
  sys.exit( 2 )

# Core of scrip
logger.info( __version__ )
logger.info( 'starting work...' )
logger.debug( 'Structure parsed from configuration file:\n\\%s\\' % config )

mbx = mboxOpen( config[ 'mailbox' ] )
if not mbx:
  logger.warning( 'cannot open mailbox' )
  sys.exit( 4 )

# XXX: prefix only for Maildir
if string.lower( config[ 'mailbox' ][ 'mboxtype' ] ) == 'maildir':
  mdirNew = os.path.join( config[ 'mailbox' ][ 'mboxpath' ], 'new' )
else:
  mdirNew = config[ 'mailbox' ][ 'mboxpath' ]

queue   = buildQueue( mbx, mdirNew, config[ 'mailbox' ][ 'dbpath' ],  \
                      config[ 'rules' ] )
if queue == False:
  logger.warn( 'problems with building queue' )
  sys.exit( 3 )
elif queue == []:
  logger.info( 'no messages in queue' )
  sys.exit( 0 )

# Queue for messages waiting for second chance
retryQueue = {}

# Process queue
for cur in queue.keys():
  try:
    sendMsg( queue[ cur ], cur,                              \
             getText( config[ 'personal' ][ 'announce' ] ),  \
             config[ 'network' ],                            \
             config[ 'personal' ] )
    logger.info( 'message processed successfully' )
  except smtplib.SMTPException, errMsg:
    logger.warn( 'cannot be processed: \'%s\' for \'%s\'' % \
        ( cur, queue[ cur ] ) )
    logger.warn( errMsg )
    logger.info( 'message will be processed later' )
    retryQueue[ cur ] = queue[ cur ]

# Last chance to be processed
for cur in retryQueue.keys():
  try:
    sendMsg( retryQueue[ cur ], cur,                              \
                  getText( config[ 'personal' ][ 'announce' ] ),  \
                  config[ 'network' ],                            \
                  config[ 'personal' ] )
    logger.info( 'message processed successfully' )
  except smtplib.SMTPException, errMsg:
    logger.warn( 'cannot be processed: \'%s\' for \'%s\'' % \
        ( cur, queue[ cur ] ) )
    logger.warn( errMsg )
    logger.warn( 'message has lost last chance to get processed' )

logger.info( 'exitting...' )

sys.exit( 0 )

# }}}

# vim: fdm=marker:tw=80:ts=6:syn=python
